@@ -1,9 +1,11 @@
 import pandas as pd
+import numpy as np
 
 import pytest
 try: import blpapi
 except ImportError: blpapi = pytest.importorskip('blpapi')
 
+from itertools import starmap
 from collections import OrderedDict
 
 from xbbg import const
@@ -19,6 +21,46 @@ BAR_TICK = blpapi.Name('barTickData')
 TICK_DATA = blpapi.Name('tickData')
 
 
+def create_request(
+        service: str,
+        request: str,
+        settings: list = None,
+        ovrds: list = None,
+        append: dict = None,
+        **kwargs,
+) -> blpapi.request.Request:
+    """
+    Create request for query
+
+    Args:
+        service: service name
+        request: request name
+        settings: list of settings
+        ovrds: list of overrides
+        append: info to be appended to request directly
+        kwargs: other overrides
+
+    Returns:
+        Bloomberg request
+    """
+    srv = conn.bbg_service(service=service, **kwargs)
+    req = srv.createRequest(request)
+
+    list(starmap(req.set, settings if settings else []))
+    if ovrds:
+        ovrd = req.getElement('overrides')
+        for fld, val in ovrds:
+            item = ovrd.appendElement()
+            item.setElement('fieldId', fld)
+            item.setElement('value', val)
+    if append:
+        for key, val in append.items():
+            vals = [val] if isinstance(val, str) else val
+            for v in vals: req.append(key, v)
+
+    return req
+
+
 def init_request(request: blpapi.request.Request, tickers, flds, **kwargs):
     """
     Initiate Bloomberg request instance
@@ -98,8 +140,9 @@ def rec_events(func, **kwargs):
     """
     timeout_counts = 0
     responses = [blpapi.Event.PARTIAL_RESPONSE, blpapi.Event.RESPONSE]
+    timeout = kwargs.pop('timeout', 500)
     while True:
-        ev = conn.bbg_session(**kwargs).nextEvent(500)
+        ev = conn.bbg_session(**kwargs).nextEvent(timeout=timeout)
         if ev.eventType() in responses:
             for msg in ev:
                 for r in func(msg=msg, **kwargs):
@@ -108,7 +151,7 @@ def rec_events(func, **kwargs):
                 break
         elif ev.eventType() == blpapi.Event.TIMEOUT:
             timeout_counts += 1
-            if timeout_counts > 10:
+            if timeout_counts > 20:
                 break
         else:
             for _ in ev:
@@ -116,7 +159,7 @@ def rec_events(func, **kwargs):
                     == SESSION_TERMINATED: break
 
 
-def process_ref(msg: blpapi.message.Message) -> dict:
+def process_ref(msg: blpapi.message.Message, **kwargs) -> dict:
     """
     Process reference messages from Bloomberg
 
@@ -126,13 +169,14 @@ def process_ref(msg: blpapi.message.Message) -> dict:
     Returns:
         dict
     """
+    kwargs.pop('(@_<)', None)
     data = None
     if msg.hasElement('securityData'):
         data = msg.getElement('securityData')
     elif msg.hasElement('data') and \
             msg.getElement('data').hasElement('securityData'):
         data = msg.getElement('data').getElement('securityData')
-    if data is None: yield {}
+    if not data: return iter([])
 
     for sec in data.values():
         ticker = sec.getElement('security').getValue()
@@ -153,7 +197,7 @@ def process_ref(msg: blpapi.message.Message) -> dict:
                 ])
 
 
-def process_hist(msg: blpapi.message.Message) -> dict:
+def process_hist(msg: blpapi.message.Message, **kwargs) -> dict:
     """
     Process historical data messages from Bloomberg
 
@@ -163,6 +207,7 @@ def process_hist(msg: blpapi.message.Message) -> dict:
     Returns:
         dict
     """
+    kwargs.pop('(>_<)', None)
     if not msg.hasElement('securityData'): return {}
     ticker = msg.getElement('securityData').getElement('security').getValue()
     for val in msg.getElement('securityData').getElement('fieldData').values():
@@ -172,7 +217,7 @@ def process_hist(msg: blpapi.message.Message) -> dict:
             ])
 
 
-def process_bar(msg: blpapi.message.Message, typ='bar') -> OrderedDict:
+def process_bar(msg: blpapi.message.Message, typ='bar', **kwargs) -> OrderedDict:
     """
     Process Bloomberg intraday bar messages
 
@@ -183,6 +228,7 @@ def process_bar(msg: blpapi.message.Message, typ='bar') -> OrderedDict:
     Yields:
         OrderedDict
     """
+    kwargs.pop('(#_#)', None)
     check_error(msg=msg)
     if typ[0].lower() == 't':
         lvls = [TICK_DATA, TICK_DATA]
@@ -208,3 +254,43 @@ def check_error(msg):
             f'{error.getElementAsString(CATEGORY)}: '
             f'{error.getElementAsString(MESSAGE)}'
         )
+
+
+def elem_value(element: conn.blpapi.Element):
+    """
+    Get value from element
+
+    Args:
+        element: Bloomberg element
+
+    Returns:
+        value
+    """
+    if element.isNull(): return None
+    value = element.getValue()
+    if isinstance(value, np.bool_): return bool(value)
+    if isinstance(value, conn.blpapi.name.Name): return str(value)
+    return value
+
+
+def earning_pct(data: pd.DataFrame, yr):
+    """
+    Calculate % of earnings by year
+    """
+    pct = f'{yr}_pct'
+    data.loc[:, pct] = np.nan
+
+    # Calculate level 1 percentage
+    data.loc[data.level == 1, pct] = \
+        100 * data.loc[data.level == 1, yr] / data.loc[data.level == 1, yr].sum()
+
+    # Calculate level 2 percentage (higher levels will be ignored)
+    sub_pct = []
+    for r, snap in data.reset_index()[::-1].iterrows():
+        if snap.level > 2: continue
+        if snap.level == 1:
+            if len(sub_pct) == 0: continue
+            data.iloc[sub_pct, data.columns.get_loc(pct)] = \
+                100 * data[yr].iloc[sub_pct] / data[yr].iloc[sub_pct].sum()
+            sub_pct = []
+        if snap.level == 2: sub_pct.append(r)
